package ru.ulstu.extractor.rule.service;

import com.fuzzylite.Engine;
import com.fuzzylite.activation.General;
import com.fuzzylite.defuzzifier.Centroid;
import com.fuzzylite.norm.s.Maximum;
import com.fuzzylite.norm.t.AlgebraicProduct;
import com.fuzzylite.norm.t.Minimum;
import com.fuzzylite.rule.Rule;
import com.fuzzylite.rule.RuleBlock;
import com.fuzzylite.term.Triangle;
import com.fuzzylite.variable.InputVariable;
import com.fuzzylite.variable.OutputVariable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;
import ru.ulstu.extractor.assessment.model.Assessment;
import ru.ulstu.extractor.rule.model.AntecedentValue;
import ru.ulstu.extractor.rule.model.DbRule;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import static java.lang.String.format;

@Service
public class FuzzyInferenceService {
    private final static Logger LOG = LoggerFactory.getLogger(FuzzyInferenceService.class);
    private final static String OUTPUT_VARIABLE_NAME = "state";
    private final static String RULE_TEMPLATE = "if %s is %s and %s is %s then "
            + OUTPUT_VARIABLE_NAME
            + " is %s";
    private final static String NO_RESULT = "Нет результата";

    private List<String> getRulesFromDb(List<DbRule> dbRules, Map<String, Double> variableValues) {
        validateVariables(variableValues, dbRules);
        return dbRules.stream().map(this::getFuzzyRule).collect(Collectors.toList());
    }

    private String getFuzzyRule(DbRule dbRule) {
        return format(RULE_TEMPLATE,
                dbRule.getFirstAntecedent().name(),
                dbRule.getFirstAntecedentValue().getAntecedentValue(),
                dbRule.getSecondAntecedent().name(),
                dbRule.getSecondAntecedentValue().getAntecedentValue(),
                dbRule.getId());
    }

    private RuleBlock getRuleBlock(Engine engine,
                                   List<DbRule> dbRules,
                                   Map<String, Double> variableValues,
                                   Map<String, Double> min,
                                   Map<String, Double> max,
                                   List<AntecedentValue> antecedentValues,
                                   List<Integer> consequentValues) {
        variableValues.forEach((key, value) -> {
            InputVariable input = new InputVariable();
            input.setName(key);
            input.setDescription("");
            input.setEnabled(true);
            double delta = antecedentValues.size() > 1
                    ? (max.get(key) - min.get(key)) / (antecedentValues.size() - 1)
                    : (max.get(key) - min.get(key));
            input.setRange(min.get(key), max.get(key));
            input.setLockValueInRange(false);
            for (int i = 0; i < antecedentValues.size(); i++) {
                input.addTerm(
                        new Triangle(
                                antecedentValues.get(i).getAntecedentValue(),
                                min.get(key) + i * delta - 0.5 * delta,
                                min.get(key) + i * delta + delta + 0.5 * delta
                        )
                );
            }
            engine.addInputVariable(input);
        });

        OutputVariable output = new OutputVariable();
        output.setName(OUTPUT_VARIABLE_NAME);
        output.setDescription("");
        output.setEnabled(true);
        output.setRange(0, consequentValues.size() + 0.1);
        output.setAggregation(new Maximum());
        output.setDefuzzifier(new Centroid(10));
        output.setDefaultValue(Double.NaN);
        output.setLockValueInRange(false);
        for (int i = 0; i < consequentValues.size(); i++) {
            output.addTerm(new Triangle(consequentValues.get(i).toString(), i, i + 2.1));
        }
        engine.addOutputVariable(output);

        RuleBlock mamdani = new RuleBlock();
        mamdani.setName("mamdani");
        mamdani.setDescription("");
        mamdani.setEnabled(true);
        mamdani.setConjunction(new Minimum());
        //mamdani.setDisjunction(null);
        mamdani.setImplication(new AlgebraicProduct());
        mamdani.setActivation(new General());
        getRulesFromDb(dbRules, variableValues).forEach(r -> {
            LOG.info(r);
            mamdani.addRule(Rule.parse(r, engine));
        });
        return mamdani;
    }

    private Engine getFuzzyEngine() {
        Engine engine = new Engine();
        engine.setName("Git rules");
        engine.setDescription("");
        return engine;
    }

    public List<Assessment> getFuzzyInference(List<DbRule> dbRules,
                                              List<AntecedentValue> antecedentValues,
                                              Map<String, Double> variableValues,
                                              Map<String, Double> min,
                                              Map<String, Double> max) {
        Engine engine = getFuzzyEngine();
        List<Integer> consequentValues = dbRules.stream().map(DbRule::getId).collect(Collectors.toList());
        engine.addRuleBlock(getRuleBlock(engine, dbRules, variableValues, min, max, antecedentValues, consequentValues));
        Map.Entry<String, Double> consequent = getConsequent(engine, variableValues);
        if (consequent.getKey().equals(NO_RESULT)) {
            return new ArrayList<>();
        }
        return dbRules
                .stream()
                .filter(r -> r.getId().equals(Integer.valueOf(consequent.getKey())))
                .map(r -> new Assessment(r, consequent.getValue()))
                .collect(Collectors.toList());
    }


    private void validateVariables(Map<String, Double> variableValues, List<DbRule> dbDbRules) {
        for (DbRule dbRule : dbDbRules) {
            if (!variableValues.containsKey(dbRule.getFirstAntecedent().name())) {
                throw new RuntimeException(format("Переменной в правиле не задано значение (нет временного ряда): %s ",
                        dbRule.getFirstAntecedent().name()));
            }
            if (!variableValues.containsKey(dbRule.getSecondAntecedent().name())) {
                throw new RuntimeException(format("Переменной в правиле не задано значение (нет временного ряда): %s ",
                        dbRule.getSecondAntecedent().name()));
            }
        }
    }

    private Map.Entry<String, Double> getConsequent(Engine engine, Map<String, Double> variableValues) {
        OutputVariable outputVariable = engine.getOutputVariable(OUTPUT_VARIABLE_NAME);
        for (Map.Entry<String, Double> variableValue : variableValues.entrySet()) {
            InputVariable inputVariable = engine.getInputVariable(variableValue.getKey());
            inputVariable.setValue(variableValue.getValue());
        }
        engine.process();
        if (outputVariable != null) {
            LOG.info("Output: {}", outputVariable.getValue());
        }
        return (outputVariable == null || Double.isNaN(outputVariable.getValue()))
                ? Map.entry(NO_RESULT, 0.0)
                : Map.entry(outputVariable.highestMembershipTerm(outputVariable.getValue()).getName(), outputVariable.getValue());
    }
}
